/*
*	Karp-Rabin algorithm
*
*	handbook, p.31
*
*	search algorithm of O(n)
*	where n is the size of text
*
*	very smart algorithm!
*	applicable to the search of multi-patterns of same length
*
*/

#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>
#include "trans.h"
#include "misc.h"

#define DBG if(0)
#define	DBG0	if(0)

Biobuf bout;

typedef struct Hkeys Hkeys;
struct Hkeys{
	char *key;
	void *value;
	Hkeys *next;
} **hkeys;

/*	nhkey should be prime number
*	use command primes to find primes
*	primes 250 270
*	251
*	257
*	263
*	269		*/
int nhkeys =251;
int keylen;
char *sep;

int pow2[256];

/*	a simple linear hash */
int
mhash(char *s, int n)
{
	int h = 0;
	while(n-- > 0) h = (h << 1)%nhkeys + *s++;
	h %= nhkeys;
	return h<0?h+nhkeys:h;
}

int
rehash(char *s, int m, int h)
{
	/*
	*	m is strlen of the pattern
	*
	*	NB: pow() returns double	*/
	h = (((h - *(s-1) * pow2[m-1]) << 1) + *(s+m-1)) % nhkeys;
	return h<0?h+nhkeys:h;
}

int
hash(char *s)
{
	return mhash(s, strlen(s));
}


void
Hput(char *key, void *value)
{
	int h = hash(key);
	Hkeys *p = new(Hkeys);
	p->key = key;
	p->value = value;
	p->next = hkeys[h];
	hkeys[h] = p;
}

void *
Hget(char *key)
{
	Hkeys *p;
	for(p = hkeys[hash(key)];p;p = p->next)
		if(strcmp(p->key, key)==0)
			return p->value;
	return nil;
}

void *
Hpull(int h, char *text)
{
	Hkeys *p;
	char *q;
	for(p = hkeys[h];p;p = p->next){
		q = p->key;
		/* it is curious, strncmp is faster than memcmp */
		//if(memcmp(q, text, strlen(q))==0)
		if(strncmp(q, text, strlen(q))==0)
			return p->value;
	}
	return nil;
}

/*
*	tabread: read translation data from rule file
*	lines beginning with "#" are comment
*	if the first line is comment and string "unicode" or "hex" is in it,
*	these are keywords for format of the file 	*/
void
tabread(char *path)
{
	int n,m;
	int fmt = 0;
	char *p,*s;
	char *args[3];
	int min,max;
	List *q;
	List *lroot;

	n = filetolist(&lroot, path);
	nhkeys = upprime(n);
	DBG print("nhkeys=%d\n",nhkeys);
	hkeys = (Hkeys**) calloc(nhkeys, sizeof(Hkeys*));
	p = lroot->data;
	if(*p == '#'){
		if(cistrstr(p,"unicode"))
			fmt = ufmt;
		else
		if(cistrstr(p,"hex"))
			fmt = hfmt;
	}


	min = 99999;
	max = 0;
	n = 0;
	for(q = lroot; q; q = q->next){
		p = q->data;
		if(sep)
			m = split(p,args,3,*sep);
		else{
			m = tokenize(p,args, 3);
			if(*p == '#')
				continue;
		}
		if(m < 2)
			continue;
		DBG print("# %s\n",args[1]);
		s = trfmt(args[0],fmt);
		if(strlen(s) == 0)
			sysfatal("# null string in the first field\n");
		if(min > strlen(s))
			min = strlen(s);
		if(max < strlen(s))
			max = strlen(s);
		Hput(s,trfmt(args[1],fmt));
		n++;
	}
	if(n == 0)
		sysfatal("# no rule");
	if(min != max)
		sysfatal("all keys must be same length");
	keylen = min;
	nrule = n;
	DBG0 print("done\n");
}


/*	advanced: O(n)	*/

#define OUTPUT(s,n) Bwrite(&bout,s,n)

void
trans(char *text)
/*	result: buf to store the result
*	ep: size of result
*	source: source text
*	ttab: table of rule
*	n: number of rules
*	buf: working buf, the size must be >= max len of ttab[n].frstr
*	return: next writing point (not null terminated)	*/
{
	int h,m;
	char *s,*t;
	char *end = text + strlen(text) - keylen;
	m = keylen;
	h = -1;
	for(t = text; t <= end; t++){
		if(h == -1)
			h = mhash(t,m);
		else{
			h = rehash(t,m,h);
			DBG print("%d %d\n",h,mhash(t,m));
		}
		s = Hpull(h,t);
		if(s){ /* matched */
			OUTPUT(s, strlen(s));
			t += m - 1; /* NB: next t++ */
			h = -1;
			continue;
		}
		OUTPUT(t, 1);
	}
	OUTPUT(t, strlen(t));
}

/*
*	translate() shows the usage example of mmtrans()
*	trans() is designed for memory to stream conversion
*/
void
translate(int fd)
{
	Biobuf bin;
	char *source;
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("input not open: %r");
	while((source = Brdstr(&bin,'\n',1))){ /* assign = */
		trans(source);
		Bwrite(&bout,"\n",1);
	}
	Bflush(&bout);
}

void
usage(void)
{
	fprint(2,
	"usage: KR_trans [-F sep] rulefile [file ....]\n"
	);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int i, fd;
	ARGBEGIN{
	case 'F': sep = EARGF(usage());
		break;
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	Binit(&bout, 1, OWRITE);

	tabread(*argv++);

	pow2[0] = 1;
	for(i = 1; i < 256; i++)
		pow2[i] = pow2[i-1]*2 % nhkeys;

	if(*argv == nil){
		translate(0);
		exits(nil);
	}

	while(*argv){
		fd = open(*argv, OREAD);
		translate(fd);
		argv++;
	}
	exits(nil);
}
